module Main where

import Data.List 
import Control.Monad

{-
* 1. Existem cinco casas.
* 2. O inglês mora na casa vermelha.
* 3. O Sueco tem um cachorro.
* 4. O dinamarquês bebe chá.
* 5. A casa verde está imediatamente a esquerda da casa branca.
* 6. Eles bebem café na casa verde.
* 7. O homem que fuma Pall Mall tem um pássaro.
* 8. Na casa amarela eles fumam Dunhill.
* 9. Na casa do meio eles bebem leite.
* 10. O norueguês mora na primeira casa.
* 11. O homem que fuma Blend mora na casa ao lado da casa com gatos.
* 12. Em uma casa ao lado da casa com um cavalo, eles fumam Dunhill.
* 13. O homem que fuma Blue Master bebe cerveja.
* 14. O Alemão fuma Prince.
* 15. O norueguês mora na casa ao lado da casa azul.
* 16. Eles bebem água na casa ao lado da casa de quem fuma Blend.

Quem é o dono da zebra?
  -}

-- Propriedades
data Cor = Vermelho | Verde | Branca | Amarela | Azul
         deriving (Eq, Enum, Bounded, Show)
data Bebida = Cha | Cafe | Leite | Cerveja | Agua
         deriving (Eq, Enum, Bounded)
data Cigarro = PallMall | Dunhill | Blend | BlueMaster | Prince
         deriving (Show, Eq, Enum, Bounded)
data Animal = Cachorro | Passaro | Gato | Cavalo | Zebra
         deriving (Eq, Enum, Bounded)
data Nacionalidade = Ingles | Dinamarques | Noruegues | Alemao | Sueco
         deriving (Eq, Enum, Bounded)

data Casa = Casa Cor Bebida Cigarro Animal Nacionalidade
          deriving Eq

instance Show Casa where
  show (Casa c b ci a n) = show c ++ show b ++ show ci ++ show a ++ show n
    {-
instance Show Cor where
  show Vermelho = "\ESC[31m"
  show Branca = "\ESC[37m"
  show Verde = "\ESC[32m"
  show Azul = "\ESC[34m"
  show Amarela  = "\ESC[33m"
-}
instance Show Bebida where
  show Cha = "🍵"
  show Cafe = "☕"
  show Leite = "🥛"
  show Cerveja = "🍺"
  show Agua = "🥤"

instance Show Animal where
  show Cachorro = "🐕"
  show Passaro = "🐦"
  show Gato = "🐈"
  show Cavalo = "🐎"
  show Zebra = "🦓" 

instance Show Nacionalidade where
  show Ingles = "🇬🇧"
  show Sueco = "🇸🇪"
  show Dinamarques = "🇩🇰"
  show Noruegues = "🇳🇴"
  show Alemao = "🇩🇪"

type Solucao = [Casa]

perms :: (Bounded a, Enum a) => [[a]]
perms = permutations [minBound .. maxBound]

eh :: Eq a => [a] -> a -> Int
-- \(a,b) -> b==x)  ==  ((==x).snd)
eh xs x = fst $ head $ filter (\(a,b) -> b==x ) $ zip [0..] xs

aEsquerdaDe, aoLadoDe :: Int -> Int -> [()]
aEsquerdaDe x y = guard $ x == y-1
aoLadoDe x y = guard $ abs (x-y) == 1

onde :: Int -> Int -> [()]
onde x y = guard $ x == y

primeira 0 = guard True
primeira x = guard False

meio 2 = guard True
meio _ = guard False

zipCasas [] [] [] [] [] = []
zipCasas (c:cs) (n:ns) (b:bs) (a:as) (ci:cis) = Casa c b a n ci : zipCasas cs ns bs as cis


solve :: [Solucao]
solve = do
  cores <- perms
  (cores `eh` Verde) `aEsquerdaDe` (cores `eh` Branca)
  
  nacoes <- perms
  primeira (nacoes `eh` Noruegues)
  onde (cores `eh` Vermelho) (nacoes `eh` Ingles)
  (nacoes `eh` Noruegues) `aoLadoDe` (cores `eh` Azul)

  bebidas <- perms
  onde (nacoes `eh` Dinamarques) (bebidas `eh` Cha)
  onde (cores `eh` Verde) (bebidas `eh` Cafe)
  meio (bebidas `eh` Leite)

  animais <- perms
  onde (nacoes `eh` Sueco) (animais `eh` Cachorro)

  cigarros <- perms
  onde (cigarros `eh` PallMall) (animais `eh` Passaro)
  onde (cores `eh` Amarela) (cigarros `eh` Dunhill)
  (cigarros `eh` Blend) `aoLadoDe` (animais `eh` Gato)
  (cigarros `eh` Dunhill) `aoLadoDe` (animais `eh` Cavalo)
  onde (cigarros `eh` BlueMaster) (bebidas `eh` Cerveja)
  onde (nacoes `eh` Alemao) (cigarros `eh` Prince)
  (cigarros `eh` Blend) `aoLadoDe` (bebidas `eh` Agua)

  return $ zipCasas cores animais bebidas cigarros nacoes 

main :: IO ()
main = do
  mapM_ (mapM_ print) $ solve
